// Copyright 2019 terrier989@gmail.com.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
// http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

import 'package:built_value/built_value.dart';
import 'package:built_value/serializer.dart';

part 'email_address.g.dart';

/// An email address.
class EmailAddress {
  static const Serializer serializer = _EmailAddressSerializer();

  static final _regExp = RegExp('.+@[^.]+(.[^.])+');

  final String value;

  const EmailAddress(this.value);

  /// Return an optional string between  '~' and '@'.
  String get box {
    final s = userAndBox;
    if (s == null) {
      return null;
    }
    final i = s.indexOf('~');
    if (i < 0) {
      return null;
    }
    return value.substring(i + 1);
  }

  /// Returns everything after '@'.
  String get domain {
    final i = value.lastIndexOf('@');
    return value.substring(i + 1);
  }

  @override
  int get hashCode => value.hashCode;

  bool get isValid => _regExp.hasMatch(value);

  /// Returns a string before the first '~' or '@'.
  String get user {
    final s = userAndBox;
    if (s == null) {
      return null;
    }
    final i = s.indexOf('~');
    if (i < 0) {
      return null;
    }
    return value.substring(0, i);
  }

  /// Returns everything before '@'.
  String get userAndBox {
    final i = value.lastIndexOf('@');
    if (i < 0) {
      return null;
    }
    return value.substring(0, i);
  }

  @override
  bool operator ==(other) => other is EmailAddress && value == other.value;

  @override
  String toString() => value;
}

/// Searches instances of [EmailAddress].
abstract class EmailAddressQuery
    implements Built<EmailAddressQuery, EmailAddressQueryBuilder> {
  static Serializer<EmailAddressQuery> get serializer =>
      _$emailAddressQuerySerializer;

  factory EmailAddressQuery.build(
          [void Function(EmailAddressQueryBuilder b) updates]) =
      _$EmailAddressQuery;

  EmailAddressQuery._();

  @nullable
  String get value;
}

class _EmailAddressSerializer implements PrimitiveSerializer<EmailAddress> {
  const _EmailAddressSerializer();

  @override
  Iterable<Type> get types => const [EmailAddress];

  @override
  String get wireName => 'phone_number';

  @override
  EmailAddress deserialize(Serializers serializers, Object serialized,
      {FullType specifiedType = FullType.unspecified}) {
    if (serialized is String) {
      return EmailAddress(serialized);
    } else {
      throw ArgumentError.value(serialized);
    }
  }

  @override
  Object serialize(Serializers serializers, EmailAddress object,
      {FullType specifiedType = FullType.unspecified}) {
    return object.value;
  }
}
