// Copyright 2019 terrier989@gmail.com.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
// http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

import 'package:built_collection/built_collection.dart';
import 'package:built_value/built_value.dart';
import 'package:built_value/serializer.dart';
import 'package:dent/core.dart';
import 'package:dent/world.dart';

part 'post.g.dart';

/// Some shared content such as a blog post, a comment, or a like.
abstract class Post extends Object
    with EntityMixin
    implements Built<Post, PostBuilder> {
  static Serializer<Post> get serializer => _$postSerializer;

  factory Post({
    Text body,
    BuiltList<Post> comments,
    BuiltList<Post> likes,
    Post parentPost,
    Post rootPost,
    String typeName,
  }) = _$Post._;

  factory Post.build([void Function(PostBuilder b) updates]) = _$Post;

  Post._();

  /// Attachments.
  @nullable
  BuiltList<Post> get attachments;

  /// Body text of the post. May be formatted in s.
  @nullable
  Text get body;

  /// Comments.
  @nullable
  BuiltList<Post> get comments;

  /// Likes.
  @nullable
  BuiltList<Post> get likes;

  /// Locale associated with the post.
  String get localeName;

  /// If this is a comment or attachment, the parent post.
  @nullable
  Post get parentPost;

  /// Place associated with post.
  @nullable
  Place get place;

  /// If this is a comment or attachment, the root post.
  Post get rootPost;

  /// A short summary.
  @nullable
  Text get summary;

  /// Title of the post.
  @nullable
  Text get title;

  /// Type of the post.
  ///
  /// Some recommended values:
  ///   * "post" - A blog post, social media post, etc.
  ///   * "comment" - A comment on a post.
  ///   * "like" - A reaction to a post.
  @nullable
  String get typeName;

  /// URI associated with the post.
  @nullable
  String get uri;
}

/// Searches instances of [Post].
abstract class PostQuery extends Object
    with EntityQueryMixin
    implements Built<PostQuery, PostQueryBuilder> {
  static Serializer<PostQuery> get serializer => _$postQuerySerializer;

  factory PostQuery({
    StringQuery typeName,
    TextQuery body,
    ListQuery<PostQuery> comments,
    ListQuery<PostQuery> likes,
    PostQuery parentPost,
    PostQuery rootPost,
    TextQuery summary,
  }) = _$PostQuery._;

  factory PostQuery.build([void Function(PostQueryBuilder b) updates]) =
      _$PostQuery;

  PostQuery._();

  /// Attachments.
  @nullable
  ListQuery<PostQuery> get attachments;

  /// Body text of the post. May be formatted in s.
  @nullable
  TextQuery get body;

  /// Comments.
  @nullable
  ListQuery<PostQuery> get comments;

  /// Likes.
  @nullable
  ListQuery<PostQuery> get likes;

  /// Locale associated with the post.
  StringQuery get localeName;

  /// If this is something like a comment, the parent post.
  @nullable
  PostQuery get parentPost;

  /// Place associated with post.
  @nullable
  PlaceQuery get place;

  /// If this is something like a comment, the root post.
  PostQuery get rootPost;

  /// A short summary.
  @nullable
  TextQuery get summary;

  /// Title of the post.
  @nullable
  TextQuery get title;

  /// Type of the post.
  ///
  /// Some recommended values:
  ///   * "post"
  ///   * "comment"
  ///   * "like"
  @nullable
  StringQuery get typeName;

  /// URI associated with the post.
  @nullable
  StringQuery get uri;
}
