// Copyright 2019 terrier989@gmail.com.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
// http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

import 'package:built_value/built_value.dart';
import 'package:built_value/serializer.dart';
import 'package:database/database.dart';
import 'package:dent/account.dart';
import 'package:dent/core.dart';

part 'meta.g.dart';

/// Metadata about an [Entity].
abstract class Meta implements Built<Meta, MetaBuilder> {
  static Serializer<Meta> get serializer => _$metaSerializer;

  factory Meta({
    bool isHistorical = false,
    Document document,
    Account owner,
    DateTime updatedAt,
  }) {
    return _$Meta._(
      isHistorical: isHistorical,
      document: document,
      owner: owner,
      updatedAt: updatedAt,
    );
  }

  factory Meta.build([void Function(MetaBuilder b) updates]) = _$Meta;

  Meta._();

  /// Additional data.
  @nullable
  Map<String, Object> get additionalFields;

  /// Database row.
  @nullable
  Document get document;

  /// True if this is some past version of the entity.
  ///
  /// By default, past versions are not retained.
  @nullable
  bool get isHistorical;

  /// The account this entity belongs to.
  ///
  /// By default:
  ///   * Only the account can see the entity.
  ///   * When the account is deleted, the entity is deleted too.
  @nullable
  Account get owner;

  /// Timestamp of this version of the entity.
  @nullable
  DateTime get updatedAt;
}

/// A query for [Meta].
abstract class MetaQuery implements Built<MetaQuery, MetaQueryBuilder> {
  static Serializer<MetaQuery> get serializer => _$metaQuerySerializer;

  factory MetaQuery({
    bool isHistorical,
    AccountQuery owner,
    DateTimeQuery updatedAt,
  }) = _$MetaQuery._;

  factory MetaQuery.build([void Function(MetaQueryBuilder b) updates]) =
      _$MetaQuery;

  MetaQuery._();

  /// True if this is some past version of the entity.
  ///
  /// By default, past versions are not retained.
  @nullable
  bool get isHistorical;

  /// The account this entity belongs to.
  ///
  /// By default:
  ///   * Only the account can see the entity.
  ///   * When the account is deleted, the entity is deleted too.
  @nullable
  AccountQuery get owner;

  /// Timestamp of this version of the entity.
  @nullable
  DateTimeQuery get updatedAt;
}
