// Copyright 2019 terrier989@gmail.com.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
// http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

import 'package:built_value/built_value.dart';
import 'package:built_value/serializer.dart';
import 'package:dent/core.dart';

part 'money.g.dart';

/// An amount of money in some currency.
///
/// You can query values with [MoneyQuery].
///
// TODO: Use integer amounts
class Money {
  static const Serializer serializer = _MoneySerializer();

  final String currency;

  final double amount;

  const Money(this.currency, this.amount)
      : assert(currency != null),
        assert(amount != null);

  /// A shorthand for US dollars.
  const Money.usd(double amount) : this('USD', amount);

  /// A shorthand for euros.
  const Money.eur(double amount) : this('EUR', amount);

  /// Adds an amount.
  ///
  /// The method checks that the currencies are the same.
  Money operator +(Money other) {
    if (currency != other.currency) {
      throw ArgumentError('Currencies dont\'t match');
    }
    return Money(currency, amount + other.amount);
  }

  /// Subtracts an amount.
  ///
  /// The method checks that the currencies are the same.
  Money operator -(Money other) {
    if (currency != other.currency) {
      throw ArgumentError('Currencies dont\'t match');
    }
    return Money(currency, amount + other.amount);
  }

  /// Multiplies the amount.
  Money scale(double scale) {
    return Money(currency, amount * scale);
  }

  @override
  int get hashCode => amount.hashCode ^ currency.hashCode;

  @override
  bool operator ==(other) =>
      other is Money && amount == other.amount && currency == other.currency;

  @override
  String toString() => '$currency $amount';
}

abstract class MoneyQuery
    implements PrimitiveQuery, Built<MoneyQuery, MoneyQueryBuilder> {
  static Serializer<MoneyQuery> get serializer => _$moneyQuerySerializer;

  factory MoneyQuery([void Function(MoneyQueryBuilder b) updates]) =
      _$MoneyQuery;

  MoneyQuery._();

  Money get max;

  Money get min;
}

class _MoneySerializer implements PrimitiveSerializer<Money> {
  const _MoneySerializer();

  @override
  Iterable<Type> get types => const [Money];

  @override
  String get wireName => 'money';

  @override
  Money deserialize(Serializers serializers, Object serialized,
      {FullType specifiedType = FullType.unspecified}) {
    if (serialized is String) {
      final i = serialized.indexOf(' ');
      final currency = serialized.substring(0, i);
      final amount = double.parse(serialized.substring(i + 1));
      return Money(currency, amount);
    } else {
      throw ArgumentError.value(serialized);
    }
  }

  @override
  Object serialize(Serializers serializers, Money object,
      {FullType specifiedType = FullType.unspecified}) {
    return '${object.currency} ${object.amount}';
  }
}
