// Copyright 2019 terrier989@gmail.com.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
// http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

import 'package:built_value/built_value.dart';
import 'package:built_value/serializer.dart';
import 'package:collection/collection.dart';
import 'package:dent/core.dart';

part 'text.g.dart';

/// Textual content.
///
/// The difference over string is support for localization directives, which are
/// in lines starting with "::".
///
/// For example, a social media post could be split into English and down-under
/// English versions:
/// ```
/// :: ?en
/// Colorful flip flops
/// ::
/// :: ?en-AU ?en-NZ // Comments are also supported.
/// ::
/// Colorful thongs
/// ```
///
/// Different users seeing different version of the text can be dangerous, which
/// application developers should remember when they design user interfaces. For
/// example, a malicious user could write a nasty text in one locale and
/// non-nasty text in another locale.
///
/// Some potential strategies to deal with danger are:
///   * Always show the raw string (ignore directives).
///   * Show a localized text, but enable users to see other locales by tapping
///     the text.
class Text {
  static const Serializer serializer = _TextSerializer();

  static List<String> defaultLocaleNames = ['en'];
  final bool _hasDirectives;

  final String _value;

  const Text(String value)
      : _hasDirectives = false,
        _value = value;

  const Text.fromRaw(String value)
      : _hasDirectives = true,
        _value = value;

  @override
  int get hashCode => raw.hashCode;

  /// Raw string.
  String get raw {
    if (_hasDirectives) {
      return _value;
    }
    // Escape lines starting with '::' as ':::'.
    var value = _value;
    if (value.startsWith('::')) {
      value = value = ':$value';
    }
    return value.replaceAll('\n::', ':::');
  }

  /// Returns all sections.
  Iterable<TextSection> get sections sync* {
    if (!_hasDirectives || !_value.contains('::')) {
      yield (TextSection(const <String>[], _value));
      return;
    }
    var directives = <String>[];
    var sb = StringBuffer();
    var previousLineWasDirective = true;
    for (var line in _value.split('\n')) {
      if (line.startsWith('::')) {
        if (line.startsWith(':::')) {
          //
          // An escaped ordinary line
          //
          if (previousLineWasDirective) {
            previousLineWasDirective = false;
          } else {
            sb.write('\n');
          }
          sb.write(line.substring(1));
        } else {
          //
          // A directive line
          //
          if (!previousLineWasDirective) {
            if (directives.isNotEmpty || sb.isNotEmpty) {
              yield (TextSection(directives, sb.toString()));
            }
            sb.clear();
            previousLineWasDirective = true;
            directives = <String>[];
          }
          line = line.substring(2);

          // Support comments
          final i = line.indexOf('//');
          if (i >= 0) {
            line = line.substring(0, i);
          }
          directives.addAll(line.split(' ').where((d) => d.isNotEmpty));
        }
      } else {
        //
        // An ordinary line
        //
        if (previousLineWasDirective) {
          previousLineWasDirective = false;
        } else {
          sb.write('\n');
        }
        sb.write(line);
      }
    }
    yield (TextSection(directives, sb.toString()));
  }

  @override
  bool operator ==(other) => other is Text && raw == other.raw;

  /// Like [raw], but directives have been trimmed.
  String allSections({String separator = '\n'}) {
    return sections.map((s) => s.content).join(separator);
  }

  /// Trims whitespace and gets only the first line.
  String firstLine({List<String> localeNames}) {
    final s = get(localeNames: localeNames).trim();
    final i = s.indexOf('\n');
    return i < 0 ? s : s.substring(0, i);
  }

  /// Returns the best localization.
  ///
  /// If [localeNames] is null, uses [defaultLocaleNames].
  String get({List<String> localeNames}) {
    if (localeNames == null) {
      localeNames = defaultLocaleNames;
    } else {
      localeNames = [...localeNames, ...defaultLocaleNames, 'en'];
    }

    // Get map
    final map = localizationsByName();
    if (map.isEmpty) {
      return '';
    }

    // Locales in the order of preference
    for (var localeName in localeNames) {
      localeName ??= '';

      // Do we have an exact match?
      final exactMatch = map[localeName];
      if (exactMatch != null) {
        return exactMatch;
      }

      // See whether one or more keys starts with the locale.
      // For example, if locale is 'en', then key 'en-US' could be used.
      // Choose alphabetically the first key.
      String bestKey;
      for (var key in map.keys) {
        if (localeName.startsWith(key) &&
            localeName.startsWith('-', key.length)) {
          if (bestKey == null || key.length > bestKey.length) {
            bestKey = key;
          }
        }
        if (key.startsWith(localeName) &&
            key.startsWith('-', localeName.length)) {
          if (bestKey == null || key.compareTo(bestKey) < 0) {
            bestKey = key;
          }
        }
      }
      if (bestKey != null) {
        return map[bestKey];
      }
    }

    // Choose alphabetically the first key.
    final key = map.keys.reduce((a, b) => a.compareTo(b) < 0 ? a : b);
    return map[key];
  }

  /// Returns all localizations.
  Map<String, String> localizationsByName({String defaultLocale}) {
    final map = <String, String>{};
    for (var section in sections) {
      final directives = section.directives;
      if (directives.isEmpty) {
        map[defaultLocale ?? ''] = section.content;
      } else {
        for (var directive in directives) {
          if (directive.startsWith('?')) {
            map[directive.substring(1)] = section.content;
          }
        }
      }
    }
    return map;
  }

  @override
  String toString() {
    if (!_hasDirectives || !raw.contains('::')) {
      return 'Text("${_escape(_value)}")';
    }
    return 'Text.withDirectives("${_escape(_value)}")';
  }

  static String _escape(String s) {
    return s.replaceAll(r'\', r'\\').replaceAll('"', r'\"');
  }
}

abstract class TextQuery
    implements PrimitiveQuery, Built<TextQuery, TextQueryBuilder> {
  static Serializer<TextQuery> get serializer => _$textQuerySerializer;

  factory TextQuery([void Function(TextQueryBuilder b) updates]) = _$TextQuery;

  TextQuery._();
}

/// A section in [Text].
class TextSection {
  /// Directives before the content.
  final List<String> directives;

  /// Content.
  final String content;

  TextSection(this.directives, this.content)
      : assert(directives != null),
        assert(content != null);

  @override
  int get hashCode => content.hashCode;

  @override
  bool operator ==(other) =>
      other is TextSection &&
      content == other.content &&
      const ListEquality().equals(directives, other.directives);

  @override
  String toString() {
    if (directives.isEmpty) {
      return content;
    }
    return ':: ${directives.join(' ')}\n$content';
  }
}

class _TextSerializer implements PrimitiveSerializer<Text> {
  const _TextSerializer();

  @override
  Iterable<Type> get types => const [Text];

  @override
  String get wireName => 'phone_number';

  @override
  Text deserialize(Serializers serializers, Object serialized,
      {FullType specifiedType = FullType.unspecified}) {
    if (serialized is String) {
      return Text.fromRaw(serialized);
    } else {
      throw ArgumentError.value(serialized);
    }
  }

  @override
  Object serialize(Serializers serializers, Text object,
      {FullType specifiedType = FullType.unspecified}) {
    return object.raw;
  }
}
