// Copyright 2019 terrier989@gmail.com.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
// http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

import 'package:built_value/built_value.dart';
import 'package:built_value/serializer.dart';
import 'package:dent/core.dart';

part 'place.g.dart';

/// A place such as a country, a city, or a business address.
///
/// Rows can be queried with [PlaceQuery].
abstract class Place extends Object
    with EntityMixin
    implements Built<Place, PlaceBuilder> {
  static Serializer<Place> get serializer => _$placeSerializer;

  factory Place({
    String countryCode,
    String stateCode,
    Text cityName,
    GeoPoint geoPoint,
    Text input,
  }) = _$Place._;

  factory Place.build([void Function(PlaceBuilder b) updates]) = _$Place;

  Place._();

  /// Apartment number
  Text get apartmentNumber;

  /// City name.
  @nullable
  Text get cityName;

  /// Two-letter country code.
  @nullable
  String get countryCode;

  /// Name of country or country-like place.
  @nullable
  Text get countryName;

  /// Geographic coordinates.
  @nullable
  GeoPoint get geoPoint;

  /// Unspecified input.
  @nullable
  Text get input;

  /// Whether this is a disputed categorization.
  ///
  /// Some places are claimed by multiple parties. Applications
  /// may want to use .
  @nullable
  bool get isDisputed;

  /// Postal code.
  @nullable
  String get postalCode;

  /// State code.
  @nullable
  String get stateCode;

  /// Street address.
  @nullable
  Text get streetAddress;

  ///  Name of subcountry or subcountry-like place.
  @nullable
  Text get subcountryName;

  /// One-line subtitle.
  Text get subtitle;

  /// One-line title.
  @nullable
  Text get title;

  /// Type of the place.
  @nullable
  String get type;
}

/// Searches instances of [Place].
abstract class PlaceQuery extends Object
    with EntityQueryMixin
    implements Built<PlaceQuery, PlaceQueryBuilder> {
  static Serializer<PlaceQuery> get serializer => _$placeQuerySerializer;

  factory PlaceQuery({
    TextQuery cityName,
    StringQuery countryCode,
    GeoPointQuery geoPoint,
    StringQuery stateCode,
    TextQuery streetAddress,
  }) = _$PlaceQuery._;

  factory PlaceQuery.build([void Function(PlaceQueryBuilder b) updates]) =
      _$PlaceQuery;

  PlaceQuery._();

  @nullable
  TextQuery get cityName;

  @nullable
  StringQuery get countryCode;

  @nullable
  GeoPointQuery get geoPoint;

  @nullable
  TextQuery get streetAddress;

  @nullable
  StringQuery get stateCode;

  @nullable
  TextQuery get title;
}
